# -*- coding: utf-8 -*-
"""
Import this module to add the validation tween to your pyramid app.
"""
import pyramid

from .api import build_swagger_20_swagger_schema_views
from .api import register_api_doc_endpoints
from .ingest import get_swagger_schema
from .ingest import get_swagger_spec
from .tween import get_swagger_versions
from .tween import SWAGGER_12
from .tween import SWAGGER_20


def includeme(config):
    """
    :type config: :class:`pyramid.config.Configurator`
    """
    settings = config.registry.settings
    swagger_versions = get_swagger_versions(settings)

    # for rendering /swagger.yaml
    config.add_renderer(
        'yaml', 'pyramid_swagger.api.YamlRendererFactory',
    )

    # Add the SwaggerSchema to settings to make it available to the validation
    # tween and `register_api_doc_endpoints`
    settings['pyramid_swagger.schema12'] = None
    settings['pyramid_swagger.schema20'] = None

    # Store under two keys so that 1.2 and 2.0 can co-exist.
    if SWAGGER_12 in swagger_versions:
        settings['pyramid_swagger.schema12'] = get_swagger_schema(settings)

    if SWAGGER_20 in swagger_versions:
        settings['pyramid_swagger.schema20'] = get_swagger_spec(settings)

    config.add_tween(
        "pyramid_swagger.tween.validation_tween_factory",
        under=pyramid.tweens.EXCVIEW
    )

    if settings.get('pyramid_swagger.enable_api_doc_views', True):
        if SWAGGER_12 in swagger_versions:
            register_api_doc_endpoints(
                config,
                settings['pyramid_swagger.schema12'].get_api_doc_endpoints())

        if SWAGGER_20 in swagger_versions:
            register_api_doc_endpoints(
                config,
                build_swagger_20_swagger_schema_views(config),
                base_path=settings.get('pyramid_swagger.base_path_api_docs', ''))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import functools
import logging
import re
import sys
from collections import namedtuple
from contextlib import contextmanager

import bravado_core
import jsonschema.exceptions
import simplejson
from bravado_core.exception import SwaggerMappingError
from bravado_core.formatter import SwaggerFormat  # noqa
from bravado_core.request import IncomingRequest
from bravado_core.request import unmarshal_request
from bravado_core.response import get_response_spec
from bravado_core.response import OutgoingResponse
from pyramid.interfaces import IRoutesMapper
from pyramid.settings import asbool
from pyramid.settings import aslist

from pyramid_swagger.exceptions import PathNotFoundError
from pyramid_swagger.exceptions import RequestValidationError
from pyramid_swagger.exceptions import ResponseValidationError
from pyramid_swagger.model import PathNotMatchedError


log = logging.getLogger(__name__)


SWAGGER_12 = '1.2'
SWAGGER_20 = '2.0'
DEFAULT_SWAGGER_VERSIONS = [SWAGGER_20]
SUPPORTED_SWAGGER_VERSIONS = [SWAGGER_12, SWAGGER_20]


DEFAULT_EXCLUDED_PATHS = [
    r'^/static/?',
    r'^/api-docs/?',
    r'^/swagger.(json|yaml)',
]


class Settings(namedtuple(
    'Settings',
    [
        'swagger12_handler',
        'swagger20_handler',
        'validate_request',
        'validate_response',
        'validate_path',
        'exclude_paths',
        'exclude_routes',
        'prefer_20_routes',
    ]
)):

    """A settings object for configurable options.

    :param swagger12_handler: a :class:`SwaggerHandler` for v1.2 or None
    :param swagger20_handler: a :class:`SwaggerHandler` for v2.0 or None
    :param validate_swagger_spec: check Swagger files for correctness.
    :param validate_request: check requests against Swagger spec.
    :param validate_response: check responses against Swagger spec.
    :param validate_path: check if request path is in schema. If disabled
        and path not found in schema, request / response validation is skipped.
    :param exclude_paths: list of paths (in regex format) that should be
        excluded from validation.
    :rtype: namedtuple
    :param exclude_routes: list of route names that should be excluded from
        validation.
    :param prefer_20_routes: list of route names that should be handled
        via v2.0 spec when `2.0` is in `swagger_versions`. All others will be
        handled via v1.2 spec. [i.e. Make v2.0 an opt-in feature]
    """


@contextmanager
def noop_context(request, response=None):
    yield


def _get_validation_context(registry):
    validation_context_path = registry.settings.get(
        'pyramid_swagger.validation_context_path',
    )

    if validation_context_path:
        m = re.match(
            '(?P<module_path>.*)\.(?P<contextmanager_name>.*)',
            validation_context_path,
        )
        module_path = m.group('module_path')
        contextmanager_name = m.group('contextmanager_name')

        return getattr(
            __import__(module_path, fromlist=contextmanager_name),
            contextmanager_name,
        )
    else:
        return noop_context


def get_swagger_objects(settings, route_info, registry):
    """Returns appropriate swagger handler and swagger spec schema.

    Swagger Handler contains callables that isolate implementation differences
    in the tween to handle both Swagger 1.2 and Swagger 2.0.

    Exception is made when `settings.prefer_20_routes` are non-empty and
    ['1.2', '2.0'] both are present in available swagger versions. In this
    special scenario, '2.0' spec is chosen only for requests which are listed
    in the `prefer_20_routes`. This helps in incremental migration of
    routes from v1.2 to v2.0 by making moving to v2.0 opt-in.

    :rtype: (:class:`SwaggerHandler`,
             :class:`pyramid_swagger.model.SwaggerSchema` OR
                :class:`bravado_core.spec.Spec`)
    """
    enabled_swagger_versions = get_swagger_versions(registry.settings)
    schema12 = registry.settings['pyramid_swagger.schema12']
    schema20 = registry.settings['pyramid_swagger.schema20']

    fallback_to_swagger12_route = (
        SWAGGER_20 in enabled_swagger_versions and
        SWAGGER_12 in enabled_swagger_versions and
        settings.prefer_20_routes and
        route_info.get('route') and
        route_info['route'].name not in settings.prefer_20_routes
    )
    if fallback_to_swagger12_route:
        return settings.swagger12_handler, schema12

    if SWAGGER_20 in enabled_swagger_versions:
        return settings.swagger20_handler, schema20

    if SWAGGER_12 in enabled_swagger_versions:
        return settings.swagger12_handler, schema12


def validation_tween_factory(handler, registry):
    """Pyramid tween for performing validation.

    Note this is very simple -- it validates requests, responses, and paths
    while delegating to the relevant matching view.
    """
    settings = load_settings(registry)
    route_mapper = registry.queryUtility(IRoutesMapper)

    def validator_tween(request):
        # We don't have access to this yet but let's go ahead and build the
        # matchdict so we can validate it and use it to exclude routes from
        # validation.
        route_info = route_mapper(request)
        swagger_handler, spec = get_swagger_objects(settings, route_info,
                                                    registry)

        if should_exclude_request(settings, request, route_info):
            return handler(request)

        validation_context = _get_validation_context(registry)

        try:
            op_or_validators_map = swagger_handler.op_for_request(
                request, route_info=route_info, spec=spec)
        except PathNotMatchedError as exc:
            if settings.validate_path:
                with validation_context(request):
                    raise PathNotFoundError(str(exc), child=exc)
            else:
                return handler(request)

        if settings.validate_request:
            with validation_context(request, response=None):
                request_data = swagger_handler.handle_request(
                    PyramidSwaggerRequest(request, route_info),
                    op_or_validators_map,
                )

            def swagger_data(_):
                return request_data

            request.set_property(swagger_data)

        response = handler(request)

        if settings.validate_response:
            with validation_context(request, response=response):
                swagger_handler.handle_response(response, op_or_validators_map)

        return response

    return validator_tween


class PyramidSwaggerRequest(IncomingRequest):
    """Adapter for a :class:`pyramid.request.Request` which exposes request
    data for casting and validation.

    The following properties are exposed in order to comply with the interface
    of :class:`bravado_core.request.IncomingRequest`:

        path: a dictionary of URL path parameters
        query: a dictionary of parameters from the query string
        form: a dictionary of form parameters from a POST
        headers: a dictionary of request headers
        files: a dictionary of uploaded filename to content
    """

    FORM_TYPES = [
        'application/x-www-form-urlencoded',
        'multipart/form-data',
    ]

    def __init__(self, request, route_info):
        """
        :type request: :class:`pyramid.request.Request`
        :type route_info: :class:`pyramid.urldispatch.Route`
        """
        self.request = request
        self.route_info = route_info

    @property
    def headers(self):
        return self.request.headers

    @property
    def query(self):
        """
        :rtype: dict
        """
        # The `mixed` dict will return a list if a parameter has multiple
        # values or a single primitive in the case of a single value.
        return self.request.GET.mixed()

    @property
    def path(self):
        return self.route_info.get('match') or {}

    @property
    def form(self):
        """
        :rtype: dict
        """
        # Don't read the POST dict unless the body is form encoded
        if self.request.content_type in self.FORM_TYPES:
            return self.request.POST.mixed()
        return {}

    @property
    def body(self):
        return self.json()

    @property
    def files(self):
        result = {}
        for k, v in self.request.params.mixed().items():
            if hasattr(v, 'file'):
                result[k] = v.file
        return result

    def json(self, **kwargs):
        return getattr(self.request, 'json_body', {})


class PyramidSwaggerResponse(OutgoingResponse):
    """Adapter for a :class:`pyramid.response.Response` which exposes response
    data for validation.

    The following properties are exposed in order to comply with the interface
    of :class:`bravado_core.response.OutgoingResponse`:

        content_type: a standard content-type string
        text: the response body as a string
        headers: a dictionary of response headers
    """

    def __init__(self, response):
        """
        :type response: :class:`pyramid.response.Response`
        """
        self.response = response

    @property
    def content_type(self):
        return self.response.content_type

    @property
    def headers(self):
        return self.response.headers

    @property
    def text(self):

        # Treating webob.Response carefully: first check if there's a
        # non-empty body attribute, since if not, we can short circuit
        if not self.response.body:
            return None
        # if there's a response body, try to read it as text, but carefully
        try:
            return self.response.text
        except AttributeError as prev_ex:
            # pyramid.response.Response raises AttributeError if you try to
            # read this property without setting the charset first: trap that
            # so it doesn't show up as a missing "text" attribute (and get
            # its error message swallowed by the base class)
            raise Exception(str(prev_ex))

    def json(self, **kwargs):
        return getattr(self.response, 'json_body', {})


def handle_request(request, validator_map, **kwargs):
    """Validate the request against the swagger spec and return a dict with
    all parameter values available in the request, casted to the expected
    python type.

    :param request: a :class:`PyramidSwaggerRequest` to validate
    :param validator_map: a :class:`pyramid_swagger.load_schema.ValidatorMap`
        used to validate the request
    :returns: a :class:`dict` of request data for each parameter in the swagger
        spec
    :raises: RequestValidationError when the request is not valid for the
        swagger spec
    """
    request_data = {}
    validation_pairs = []

    for validator, values in [
        (validator_map.query, request.query),
        (validator_map.path, request.path),
        (validator_map.form, request.form),
        (validator_map.headers, request.headers),
    ]:
        values = cast_params(validator.schema, values)
        validation_pairs.append((validator, values))
        request_data.update(values)

    # Body is a special case because the key for the request_data comes
    # from the name in the schema, instead of keys in the values
    if validator_map.body.schema:
        param_name = validator_map.body.schema['name']
        validation_pairs.append((validator_map.body, request.body))
        request_data[param_name] = request.body

    validate_request(validation_pairs)

    return request_data


def load_settings(registry):
    return Settings(
        swagger12_handler=build_swagger12_handler(
            registry.settings.get('pyramid_swagger.schema12')),
        swagger20_handler=build_swagger20_handler(),
        validate_request=asbool(registry.settings.get(
            'pyramid_swagger.enable_request_validation',
            True,
        )),
        validate_response=asbool(registry.settings.get(
            'pyramid_swagger.enable_response_validation',
            True,
        )),
        validate_path=asbool(registry.settings.get(
            'pyramid_swagger.enable_path_validation',
            True,
        )),
        exclude_paths=get_exclude_paths(registry),
        exclude_routes=set(aslist(registry.settings.get(
            'pyramid_swagger.exclude_routes',
        ) or [])),
        prefer_20_routes=set(aslist(registry.settings.get(
            'pyramid_swagger.prefer_20_routes') or [])),
    )


SwaggerHandler = namedtuple('SwaggerHandler',
                            'op_for_request handle_request handle_response')


def build_swagger20_handler():
    return SwaggerHandler(
        op_for_request=get_op_for_request,
        handle_request=swaggerize_request,
        handle_response=swaggerize_response,
    )


def build_swagger12_handler(schema):
    """Builds a swagger12 handler or returns None if no schema is present.

    :type schema: :class:`pyramid_swagger.model.SwaggerSchema`
    :rtype: :class:`SwaggerHandler` or None
    """
    if schema:
        return SwaggerHandler(
            op_for_request=schema.validators_for_request,
            handle_request=handle_request,
            handle_response=validate_response,
        )


def get_exclude_paths(registry):
    """Compiles a list of paths that should not be validated against.
        :rtype: list of compiled validation regexes
    """
    # TODO(#63): remove deprecated `skip_validation` setting in v2.0.
    regexes = registry.settings.get(
        'pyramid_swagger.skip_validation',
        registry.settings.get(
            'pyramid_swagger.exclude_paths',
            DEFAULT_EXCLUDED_PATHS
        )
    )

    # being nice to users using strings :p
    if not isinstance(regexes, list) and not isinstance(regexes, tuple):
        regexes = [regexes]

    return [re.compile(r) for r in regexes]


def is_swagger_documentation_route(route_info):
    if not route_info:
        return False

    route = route_info.get('route')
    if not route:
        return False
    return route.name.startswith('pyramid_swagger.swagger20.api_docs.')


def should_exclude_request(settings, request, route_info):
    disable_all_validation = not any((
        settings.validate_request,
        settings.validate_response,
        settings.validate_path
    ))
    return (
        disable_all_validation or
        should_exclude_path(settings.exclude_paths, request.path_info) or
        should_exclude_route(settings.exclude_routes, route_info) or
        is_swagger_documentation_route(route_info)
    )


def should_exclude_path(exclude_path_regexes, path):
    # Skip validation for the specified endpoints
    return any(r.match(path) for r in exclude_path_regexes)


def should_exclude_route(excluded_routes, route_info):
    return (
        route_info.get('route') and
        route_info['route'].name in excluded_routes
    )


def validation_error(exc_class):
    def decorator(f):
        @functools.wraps(f)
        def _validate(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except (jsonschema.exceptions.ValidationError,
                    SwaggerMappingError) as exc:
                # This will alter our stack trace slightly, but Pyramid knows
                # how to render it. And the real value is in the message
                # anyway.
                e = exc_class(str(exc), child=exc)
                e._traceback = sys.exc_info()[2]
                raise e

        return _validate

    return decorator


CAST_TYPE_TO_FUNC = {
    'integer': int,
    'float': float,
    'number': float,
    'boolean': bool,
}


def cast_request_param(param_type, param_name, param_value):
    """Try to cast a request param (e.g. query arg, POST data) from a string to
    its specified type in the schema. This allows validating non-string params.

    :param param_type: name of the type to be casted to
    :type  param_type: string
    :param param_name: param name
    :type  param_name: string
    :param param_value: param value
    :type  param_value: string
    """
    try:
        return CAST_TYPE_TO_FUNC.get(param_type, lambda x: x)(param_value)
    except ValueError:
        log.warn("Failed to cast %s value of %s to %s",
                 param_name, param_value, param_type)
        # Ignore type error, let jsonschema validation handle incorrect types
        return param_value


@validation_error(RequestValidationError)
def validate_request(validation_pairs):
    for validator, values in validation_pairs:
        validator.validate(values)


def cast_params(schema, values):
    if not schema:
        return {}

    def get_type(param_name):
        return schema['properties'].get(param_name, {}).get('type')

    return dict(
        (k, cast_request_param(get_type(k), k, v))
        for k, v in values.items()
    )


@validation_error(ResponseValidationError)
def validate_response(response, validator_map):
    """Validates response against our schemas.

    :param response: the response object to validate
    :type response: :class:`pyramid.response.Response`
    :type validator_map: :class:`pyramid_swagger.load_schema.ValidatorMap`
    """
    validator = validator_map.response

    # Short circuit if we are supposed to not validate anything.
    returns_nothing = validator.schema.get('type') == 'void'
    body_empty = response.body in (None, b'', b'{}', b'null')
    if returns_nothing and body_empty:
        return

    # Don't attempt to validate non-success responses in v1.2
    if not 200 <= response.status_code <= 203:
        return

    validator.validate(prepare_body(response))


def prepare_body(response):
    # content_type must be set to access response.text
    if not response.content_type:
        raise ResponseValidationError(
            'Response validation error: Content-Type must be set'
        )

    if 'application/json' in response.content_type:
        return simplejson.loads(response.text)
    else:
        return response.text


@validation_error(RequestValidationError)
def swaggerize_request(request, op, **kwargs):
    """
    Delegate handling the Swagger concerns of the request to bravado-core.
    Post-invocation, the Swagger request parameters are available as a dict
    named `swagger_data` on the Pyramid request.

    :type request: :class:`pyramid.request.Request`
    :type op: :class:`bravado_core.operation.Operation`
    """
    request_data = unmarshal_request(request, op)
    return request_data


@validation_error(ResponseValidationError)
def swaggerize_response(response, op):
    """
    Delegate handling the Swagger concerns of the response to bravado-core.

    :type response: :class:`pyramid.response.Response`
    :type op: :class:`bravado_core.operation.Operation`
    """
    response_spec = get_response_spec(response.status_int, op)
    bravado_core.response.validate_response(
        response_spec, op, PyramidSwaggerResponse(response))


def get_op_for_request(request, route_info, spec):
    """
    Find out which operation in the Swagger schema corresponds to the given
    pyramid request.

    :type request: :class:`pyramid.request.Request`
    :type route_info: dict (usually has 'match' and 'route' keys)
    :type spec: :class:`bravado_core.spec.Spec`
    :rtype: :class:`bravado_core.operation.Operation`
    :raises: PathNotMatchedError when a matching Swagger operation is not
        found.
    """
    # pyramid.urldispath.Route
    route = route_info['route']
    if hasattr(route, 'path'):
        route_path = route.path
        if route_path[0] != '/':
            route_path = '/' + route_path
        op = spec.get_op_for_request(request.method, route_path)
        if op is not None:
            return op
        else:
            raise PathNotMatchedError(
                "Could not find a matching Swagger "
                "operation for {0} request {1}"
                .format(request.method, request.url))
    else:
        raise PathNotMatchedError(
            "Could not find a matching route for {0} request {1}. "
            "Have you registered this endpoint with Pyramid?"
            .format(request.method, request.url))


def get_swagger_versions(settings):
    """
    Validates and returns the versions of the Swagger Spec that this pyramid
    application supports.

    :type settings: dict
    :return: list of strings. eg ['1.2', '2.0']
    :raises: ValueError when an unsupported Swagger version is encountered.
    """
    swagger_versions = set(aslist(settings.get(
        'pyramid_swagger.swagger_versions', DEFAULT_SWAGGER_VERSIONS)))

    if len(swagger_versions) == 0:
        raise ValueError('pyramid_swagger.swagger_versions is empty')

    for swagger_version in swagger_versions:
        if swagger_version not in SUPPORTED_SWAGGER_VERSIONS:
            raise ValueError('Swagger version {0} is not supported.'
                             .format(swagger_version))
    return swagger_versions



# -*- coding: utf-8 -*-
"""
Module for automatically serving /api-docs* via Pyramid.
"""
import copy
import os.path

import simplejson
import yaml
from bravado_core.spec import strip_xscope
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse

from pyramid_swagger.model import PyramidEndpoint


# TODO: document that this is now a public interface
def register_api_doc_endpoints(config, endpoints, base_path='/api-docs'):
    """Create and register pyramid endpoints to service swagger api docs.
    Routes and views will be registered on the `config` at `path`.

    :param config: a pyramid configuration to register the new views and routes
    :type  config: :class:`pyramid.config.Configurator`
    :param endpoints: a list of endpoints to register as routes and views
    :type  endpoints: a list of :class:`pyramid_swagger.model.PyramidEndpoint`
    :param base_path: the base path used to register api doc endpoints.
        Defaults to `/api-docs`.
    :type  base_path: string
    """
    for endpoint in endpoints:
        path = base_path.rstrip('/') + endpoint.path
        config.add_route(endpoint.route_name, path)
        config.add_view(
            endpoint.view,
            route_name=endpoint.route_name,
            renderer=endpoint.renderer)


def build_swagger_12_endpoints(resource_listing, api_declarations):
    """
    :param resource_listing: JSON representing a Swagger 1.2 resource listing
    :type resource_listing: dict
    :param api_declarations: JSON representing Swagger 1.2 api declarations
    :type api_declarations: dict
    :rtype: iterable of :class:`pyramid_swagger.model.PyramidEndpoint`
    """
    yield build_swagger_12_resource_listing(resource_listing)

    for name, filepath in api_declarations.items():
        with open(filepath) as input_file:
            yield build_swagger_12_api_declaration(
                name, simplejson.load(input_file))


def build_swagger_12_resource_listing(resource_listing):
    """
    :param resource_listing: JSON representing a Swagger 1.2 resource listing
    :type resource_listing: dict
    :rtype: :class:`pyramid_swagger.model.PyramidEndpoint`
    """
    def view_for_resource_listing(request):
        # Thanks to the magic of closures, this means we gracefully return JSON
        # without file IO at request time.
        return resource_listing

    return PyramidEndpoint(
        path='',
        route_name='pyramid_swagger.swagger12.api_docs',
        view=view_for_resource_listing,
        renderer='json')


def build_swagger_12_api_declaration(resource_name, api_declaration):
    """
    :param resource_name: The `path` parameter from the resource listing for
        this resource.
    :type resource_name: string
    :param api_declaration: JSON representing a Swagger 1.2 api declaration
    :type api_declaration: dict
    :rtype: :class:`pyramid_swagger.model.PyramidEndpoint`
    """
    # NOTE: This means our resource paths are currently constrained to be valid
    # pyramid routes! (minus the leading /)
    route_name = 'pyramid_swagger.swagger12.apidocs-{0}'.format(resource_name)
    return PyramidEndpoint(
        path='/{0}'.format(resource_name),
        route_name=route_name,
        view=build_swagger_12_api_declaration_view(api_declaration),
        renderer='json')


def build_swagger_12_api_declaration_view(api_declaration_json):
    """Thanks to the magic of closures, this means we gracefully return JSON
    without file IO at request time.
    """
    def view_for_api_declaration(request):
        # Note that we rewrite basePath to always point at this server's root.
        return dict(
            api_declaration_json,
            basePath=str(request.application_url),
        )
    return view_for_api_declaration


def _low_level_translate(url, is_marshaling=True):
    """
    Translate the URL string representation into a new string which could be
    used as JSON keys.
    Example: if the url is "#/definitions/data_type" it could not be directly
    injected into the JSON as a key since many parser could comply with it.
    To workaround this limitation we can re-write the url representation in a
    way that the parsers will accept it, for example "#/definitions/data_type"
    could become "|..definitions..data_type"

    :param url: string representation of an url (the expected format is
        '[{scheme}://][{host}]{path}[#{fragment}]')
    :type url: str
    :param is_marshaling: flag representing the marshaling (to JSON) or the
        un-marshaling (from JSON) operation
    :type is_marshaling: bool
    :return: a string representation of the URL which could be used into the
        JSON keys
    """
    # replacement of the scheme (valid only in un-marshaling phase)
    scheme_replacements = [  # defined as tuples (un-marshaled, marshaled)
        ('file://', 'file.'),
        ('http://', 'http.'),
        ('https://', 'https.'),
    ]
    # replacement over the whole string
    replacement_patterns = [  # defined as tuples (un-marshaled, marshaled)
        ('/', '..'),
        ('#', '|'),
    ]

    # repl_idx allow to prevent if statements over the code execution
    # instead of doing:
    #       if is_marshaling:
    #           x = x.replace(replacement[0], replacement[1])
    #       else:
    #           x = x.replace(replacement[1], replacement[0])
    # we can achieve the same doing
    #       x = x.replace(replacement[repl_idx], replacement[1-repl_idx])
    repl_idx = 0 if is_marshaling else 1
    replaced_scheme = ''
    for replacement in scheme_replacements:
        if url.startswith(replacement[repl_idx]):
            replaced_scheme = replacement[1 - repl_idx]
            url = url[len(replacement[repl_idx]):]
            break
    for replacement in replacement_patterns:
        url = url.replace(replacement[repl_idx], replacement[1 - repl_idx])
    return replaced_scheme + url


def _get_absolute_link(spec, resource_path_fragment, current_path=''):
    """
    Generate an absolute link starting form the relative path and the
    current_path information.

    :param spec: Swagger Spec
    :type spec: bravado_core.spec.Spec
    :param resource_path_fragment: path and fragment of the resource
        (the expected format is '{path}#{fragment}')
    :type resource_path_fragment: str
    :param current_path: local or remote base path used for the absolute URL
        generation. It represent the path of the resource in which is defined
        the target/reference.
    :type current_path: str
    :return: URL containing the absolute path associated to the relative_path
    """
    spec_url = urlparse(spec.origin_url)
    spec_file = os.path.abspath(spec_url.path)
    spec_dir = os.path.dirname(spec_file)

    current_path_url = urlparse(current_path)

    target_url = urlparse(resource_path_fragment)
    if len(target_url.scheme) > 0:
        target_scheme = target_url.scheme
    elif len(current_path_url.scheme) > 0:
        target_scheme = current_path_url.scheme
    else:
        target_scheme = 'file'

    if target_scheme == 'file':  # targeting a local file
        # if not absolute path, make it absolute
        if not target_url.path.startswith('/'):
            # if path is empty, then is targeting to the current file
            filename = target_url.path if len(target_url.path) > 0 \
                else os.path.basename(current_path_url.path)
            target_url = urlparse('{path}#{fragment}'.format(
                path=os.path.join(
                    spec_dir,
                    os.path.dirname(current_path_url.path),
                    filename,
                ),
                fragment=target_url.fragment,
            ))
        return target_url

    elif target_scheme in ['http', 'https']:
        remote_path = os.path.abspath(os.path.join(
            current_path_url.path if len(current_path_url.path) > 0 else '/',
            target_url.path
        ))
        return urlparse('{scheme}://{hostname}{path}#{fragment}'.format(
            scheme=target_scheme,
            hostname=target_url.hostname,
            path=remote_path,
            fragment=target_url.fragment,
        ))
    else:
        return urlparse(resource_path_fragment)


def _get_target_url(spec, target, current_path=''):
    """
    Generate a well formatted URL for the required target.
    NOTE: The method assumes that the base swagger definition file
    (swagger.{json,yaml}) is a server local file.

    The URL ({schema}://{host}{path}#{fragment}) will be:
        - fully defined for a remote resources
          The URL associated to "$ref": "http://link/path.json#/definitions/dt)
          is "http://link/path.json#/definitions/dt"
        - not fully defined for local resources. The schema and host components
          of the URL will be empty and the path will contain only the relative
          path respect to the base swagger file (swagger.{json,yaml})
    NOTE: We are not fully referencing the local resources in order to limit as
    much as possible the amount of information that are made public (ie. the
    complete server paths containing the swagger files). The limitation, of
    server related, information will make harder for an intruder to forge or
    modify the server specifications.

    :param spec: Swagger Spec
    :type spec: bravado_core.spec.Spec
    :param target: URL of the target that has to be formatted
    :type target: str
    :param current_path: local or remote base path used for the absolute URL
        generation. It represent the path of the resource in which is defined
        the target/reference.
    :type current_path: str
    :return: url associated to the target resource
    """
    if target == '':
        raise ValueError('Empty target')

    target_url = _get_absolute_link(spec, target, current_path)
    target_scheme = 'file' if target_url.scheme == '' else target_url.scheme

    if target_scheme == 'file':
        spec_dir = os.path.dirname(urlparse(spec.origin_url).path)
        # Hiding of the internal server paths information.
        # A path relative to the swagger.{json,yaml} is returned
        target_url = urlparse('{path}#{fragment}'.format(
            path=os.path.relpath(target_url.path, spec_dir),
            fragment=target_url.fragment,
        ))

    if target_scheme in ['file', 'http', 'https']:
        return target_url
    else:  # Handle the case of an unknown target scheme
        raise ValueError(
            'Invalid target: {target}'.format(target=target)
        )


def _ensure_sane_key(fragment, mar_target):
    '''
    This function patches a nasty bug with the
    which was corrupting out keys

    explanation:
        sometimes the fragment becomes:
        '/definitions/:........swagger|..definitions..A'
        and which makes the marshaled_target:
        ':........swagger|..definitions..:........swagger|..definitions..A'
        this is bad, it only comes about in a very specific case
        this function esentially checks for that.
    In the above example we want to return: ':........swagger|..definitions..A'

    see tests/sample_specs/nested_defns/swagger.yaml
    for an example spec that causes this issue.

    '''
    simple_fragment = fragment.split('/')[-1]
    path_to_key = simple_fragment.split('..')
    path = path_to_key[:-1]
    key = path_to_key[-1]
    bad_path = '..'.join(path + path + [key])
    if bad_path == mar_target:
        return '..'.join(path + [key])
    return mar_target


def _marshal_target(target_url):
    target_scheme = target_url.scheme
    target_fragment = target_url.fragment
    if len(target_url.path) > 0 and \
            target_scheme in ['', 'file', 'http', 'https']:
        marshaled_target = _low_level_translate(
            '{scheme}://{host}{path}#{fragment}'.format(
                scheme=target_url.scheme,
                host=target_url.hostname if target_url.hostname else '',
                path=target_url.path,
                fragment=target_fragment,
            ))
        return _ensure_sane_key(target_fragment, marshaled_target)
    else:
        raise ValueError(
            'Invalid target: {target}'.format(target=urlunparse(target_url))
        )


def _unmarshal_target(marshaled_target):
    if any(marshaled_target.startswith(x) for x in
           ('file.', 'http.', 'https.')):
        result = _low_level_translate(marshaled_target, is_marshaling=False)
        # Remove the file:// scheme to allow the use of relative paths
        if not result.startswith('file:///'):
            result = result.replace('file://', '')
        return result
    else:
        raise ValueError(
            'Invalid marshaled object: {target}'.format(
                target=marshaled_target
            )
        )


def is_a_swagger_definition(json_path):
    """
    The method checks if json_path could have references to Swagger definitions
    (http://swagger.io/specification/#definitionsObject).

    :param json_path: json path to check
    :type json_path: list
    :return: true if json_path could have references toward Swagger definition
    objects, false otherwise
    :type: bool
    """
    # According to the Swagger specifications only a subset of paths could
    # target a Swagger definition object.
    # The paths that could led to a Swagger definition object are:
    # '/definitions/.*', '.*/schema',  '.*/items' and '.*/allOf'
    return json_path == ['definitions'] or \
        json_path[-1] in ['schema', 'items', 'allOf']


def resolve_ref(spec, url, json_path, file_path, defs_dict):
    with spec.resolver.resolving(url) as resolved_spec_dict:
        spec_dict = copy.deepcopy(resolved_spec_dict)
        return resolve_refs(spec, spec_dict, json_path, file_path, defs_dict)


def make_resolved_ref(marshaled_target):
    return {'$ref': '#/definitions/{target}'.format(
        target=marshaled_target,
    )}


def resolve_refs(spec, val, json_path, file_path, defs_dict):
    if isinstance(val, dict):
        new_dict = {}
        for key, subval in val.items():
            if key == '$ref':
                if is_a_swagger_definition(json_path):
                    target_url = _get_target_url(spec, subval, file_path)
                    marshaled_target = _marshal_target(target_url)
                    if marshaled_target not in defs_dict:
                        defs_dict[marshaled_target] = None  # placeholder
                        # The placeholder is present to interrupt the recursion
                        # during the resolve_ref of a recursive data model
                        defs_dict[marshaled_target] = resolve_ref(
                            spec, subval, json_path, file_path, defs_dict
                        )
                    return make_resolved_ref(marshaled_target)
                # assume $ref is the only key in the dict
                return resolve_ref(
                    spec, subval, json_path, file_path, defs_dict
                )
            else:
                new_dict[key] = resolve_refs(
                    spec, subval, json_path + [key], file_path, defs_dict
                )
        return new_dict

    if isinstance(val, list):
        for index, subval in enumerate(val):
            val[index] = resolve_refs(
                spec, subval, json_path + [index], file_path, defs_dict
            )
    return val


class NodeWalker(object):
    def __init__(self):
        pass

    def walk(self, item, *args, **kwargs):
        dupe = copy.deepcopy(item)
        return self._walk(dupe, *args, **kwargs)

    def _walk(self, item, *args, **kwargs):
        if isinstance(item, list):
            return self._walk_list(item, *args, **kwargs)

        elif isinstance(item, dict):
            return self._walk_dict(item, *args, **kwargs)

        else:
            return self._walk_item(item, *args, **kwargs)

    def _walk_list(self, item, *args, **kwargs):
        for index, subitem in enumerate(item):
            item[index] = self._walk(subitem, *args, **kwargs)
        return item

    def _walk_dict(self, item, *args, **kwargs):
        for key, value in item.items():
            item[key] = self._walk_dict_item(key, value, *args, **kwargs)
        return item

    def _walk_dict_item(self, key, value, *args, **kwargs):
        return self._walk(value, *args, **kwargs)

    def _walk_item(self, value, *args, **kwargs):
        return value


def get_path_if_relative(url):
    parts = urlparse(url)

    if parts.scheme or parts.netloc:
        # only rewrite relative paths
        return

    if not parts.path:
        # don't rewrite internal refs
        return

    if parts.path.startswith('/'):
        # don't rewrite absolute refs
        return

    return parts


class NodeWalkerForRefFiles(NodeWalker):
    def walk(self, spec):
        all_refs = []

        spec_fname = spec.origin_url
        if spec_fname.startswith('file://'):
            spec_fname = spec_fname.replace('file://', '')
        spec_dirname = os.path.dirname(spec_fname)

        parent = super(NodeWalkerForRefFiles, self)
        parent.walk(spec.client_spec_dict, spec, spec_dirname, all_refs)

        all_refs = [os.path.relpath(f, spec_dirname) for f in all_refs]
        all_refs = set(all_refs)

        core_dirname, core_fname = os.path.split(spec_fname)
        all_refs.add(core_fname)

        return all_refs

    def _walk_dict_item(self, key, value, spec, dirname, all_refs):
        if key != '$ref':
            parent = super(NodeWalkerForRefFiles, self)
            return parent._walk_dict_item(key, value, spec, dirname, all_refs)

        # assume $ref is the only key in the dict
        parts = get_path_if_relative(value)
        if not parts:
            return value

        full_fname = os.path.join(dirname, parts.path)
        norm_fname = os.path.normpath(full_fname)
        all_refs.append(norm_fname)

        with spec.resolver.resolving(value) as spec_dict:
            dupe = copy.deepcopy(spec_dict)
            self._walk(dupe, spec, os.path.dirname(norm_fname), all_refs)


class NodeWalkerForCleaningRefs(NodeWalker):
    def walk(self, item, schema_format):
        parent = super(NodeWalkerForCleaningRefs, self)
        return parent.walk(item, schema_format)

    @staticmethod
    def fix_ref(ref, schema_format):
        parts = get_path_if_relative(ref)
        if not parts:
            return

        path, ext = os.path.splitext(parts.path)
        return urlunparse([
            parts.scheme,
            parts.netloc,
            '{0}.{1}'.format(path, schema_format),
            parts.params,
            parts.query,
            parts.fragment,
        ])

    def _walk_dict_item(self, key, value, schema_format):
        if key != '$ref':
            parent = super(NodeWalkerForCleaningRefs, self)
            return parent._walk_dict_item(key, value, schema_format)

        return self.fix_ref(value, schema_format) or value


class YamlRendererFactory(object):
    def __init__(self, info):
        pass

    def __call__(self, value, system):
        response = system['request'].response
        response.headers['Content-Type'] = 'application/x-yaml; charset=UTF-8'
        return yaml.dump(value).encode('utf-8')


def build_swagger_20_swagger_schema_views(config):
    settings = config.registry.settings
    if settings.get('pyramid_swagger.dereference_served_schema'):
        views = _build_dereferenced_swagger_20_schema_views(config)
    else:
        views = _build_swagger_20_schema_views(config)
    return views


def _build_dereferenced_swagger_20_schema_views(config):
    def view_for_swagger_schema(request):
        settings = config.registry.settings
        resolved_dict = settings.get('pyramid_swagger.schema20_resolved')
        if not resolved_dict:
            spec = settings['pyramid_swagger.schema20']
            spec_copy = copy.deepcopy(spec.client_spec_dict)
            spec_file_name = os.path.basename(urlparse(spec.origin_url).path)
            defs_dict = {}
            resolved_dict = resolve_refs(spec, spec_copy, ['/'],
                                         spec_file_name, defs_dict)
            if len(defs_dict) > 0:
                resolved_dict.update({'definitions': defs_dict})
            settings['pyramid_swagger.schema20_resolved'] = resolved_dict
        return resolved_dict

    for schema_format in ['yaml', 'json']:
        route_name = 'pyramid_swagger.swagger20.api_docs.{0}'\
            .format(schema_format)
        yield PyramidEndpoint(
            path='/swagger.{0}'.format(schema_format),
            view=view_for_swagger_schema,
            route_name=route_name,
            renderer=schema_format,
        )


def _build_swagger_20_schema_views(config):
    spec = config.registry.settings['pyramid_swagger.schema20']

    walker = NodeWalkerForRefFiles()
    all_files = walker.walk(spec)

    file_map = {}

    def view_for_swagger_schema(request):
        _, ext = os.path.splitext(request.path)
        ext = ext.lstrip('.')

        base_path = config.registry.settings\
            .get('pyramid_swagger.base_path_api_docs', '').rstrip('/')

        key_path = request.path_info[len(base_path):]

        actual_fname = file_map[key_path]

        with spec.resolver.resolving(actual_fname) as spec_dict:
            clean_response = strip_xscope(spec_dict)
            ref_walker = NodeWalkerForCleaningRefs()
            fixed_spec = ref_walker.walk(clean_response, ext)
            return fixed_spec

    for ref_fname in all_files:
        ref_fname_parts = os.path.splitext(ref_fname)
        for schema_format in ['yaml', 'json']:
            route_name = 'pyramid_swagger.swagger20.api_docs.{0}.{1}'\
                .format(ref_fname.replace('/', '.'), schema_format)
            path = '/{0}.{1}'.format(ref_fname_parts[0], schema_format)
            file_map[path] = ref_fname
            yield PyramidEndpoint(
                path=path,
                route_name=route_name,
                view=view_for_swagger_schema,
                renderer=schema_format,
            )
